/****************************************************************************

FILE	s_dump.c

PURPOSE 
    This file contains support functions for opcodes.

*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/

#include <stdio.h>
#include "opcodes.h"
#include "shade.h"

int output_functions = 0;

struct ops optab[] = 
{
    {oASSIGN3, "assign3", 3},
    {oASSIGN, "assign", 3},
    {oASSIGNS, "assigns",3},
    {oADDEQ3, "addeq3", 3},
    {oADDEQ, "addeq", 3},
    {oSUBEQ3, "subeq3", 3},
    {oSUBEQ, "subeq", 3},
    {oMULEQ3, "muleq3", 3},
    {oMULEQ, "muleq", 3},
    {oDIVEQ3, "diveq3", 3},
    {oDIVEQ, "diveq", 3},
    {oADD3, "add3", 1},
    {oADD, "add", 1},
    {oSUB3, "sub3", 1},
    {oSUB, "sub", 1},
    {oMUL3, "mul3", 1},
    {oMUL, "mul", 1},
    {oDIV3, "div3", 1},
    {oDIV, "div", 1},
    {oDOTP, "dotp", 1},
    {oCROSSP, "crossp", 1},
    {oNEGATE3, "negate3", 1},
    {oNEGATE, "negate", 1},
    {oPUSH, "push", 3},
    {oPUSH3, "push3", 3},
    {oPUSHARR, "pusharr", 3},
    {oPUSHARR3, "pusharr3", 3},
    {oPUSHALEN, "pushalen", 3},
    {oPUSHC, "pushc", 3},
    {oPUSHS, "pushs", 3},
    {oSIN, "sin", 1},
    {oASIN, "asin", 1},
    {oCOS, "cos", 1},
    {oACOS, "acos", 1},
    {oTAN, "tan", 1},
    {oATAN, "atan", 1},
    {oATAN2, "atan2", 1},
    {oRADIANS, "radians", 1},
    {oDEGREES, "degrees", 1},
    {oSQR, "sqr", 1},
    {oSQRT, "sqrt", 1},
    {oINVSQRT, "inversesqrt", 1},
    {oPOW, "pow", 1},
    {oEXP, "exp", 1},
    {oLOG, "log", 1},
    {oLOG2, "log2", 1},
    {oMOD, "mod", 1},
    {oABS, "abs", 1},
    {oSIGN, "sign", 1},
    {oMIN, "min", 1},
    {oMIN3, "min3", 1},
    {oMAX, "max", 1},
    {oMAX3, "max3", 1},
    {oCLAMP, "clamp", 1},
    {oCLAMP3, "clamp3", 1},
    {oCEIL, "ceil", 1},
    {oFLOOR, "floor", 1},
    {oROUND, "round", 1},
    {oSTEP, "step", 1},
    {oSMOOTHSTEP, "smoothstep", 1},
    {oTRANSFORM, "transform", 5},
    {oNTRANSFORM, "ntransform", 5},
    {oVTRANSFORM, "vtransform", 5},
    {oLENGTH, "length", 1},
    {oDISTANCE, "distance", 1},
    {oAREA, "area", 1},
    {oNORMALIZE, "normalize", 1},
    {oMIX, "mix", 1},
    {oMIX3, "mix", 1},
    {oNOISE, "noise", 3},
    {oNOISE3, "noise3", 3},
    {oAMBIENT, "ambient", 1},
    {oDIFFUSE, "diffuse", 1},
    {oPHONG, "phong", 1},
    {oSPECULAR, "specular", 1},
    {oSPECULARBRDF, "specularbrdf", 1},
    {oTRACE, "trace", 1},
    {oFACEFORWARD, "faceforward", 1},
    {oEQ3, "eq3", 3},
    {oEQ, "eq", 3},
    {oEQS, "eqs", 3},
    {oNE3, "ne3", 3},
    {oNE, "ne", 3},
    {oNES, "nes", 3},
    {oLT, "lt", 3},
    {oGT, "gt", 3},
    {oLE, "le", 3},
    {oGE, "ge", 3},
    {oJMP, "jmp", 3},
    {oRET, "ret", 1},
    {oPRINTF, "printf", 5},
    {oSPLINE, "spline", 3},
    {oSPLINE3, "spline3", 3},
    {oEND, "end", 1},
    {oENDINIT, "endinit", 1},
    {oCALL, "call", 5},
    {oPOP, "pop", 1},
    {oPOP3, "pop3", 1},
    {oCTRANSFORM, "ctransform", 5},
    {oCOMP, "comp", 1},
    {oXCOMP, "xcomp", 1},
    {oYCOMP, "ycomp", 1},
    {oZCOMP, "zcomp", 1},
    {oSETCOMP, "setcomp", 3},
    {oSETXCOMP, "setxcomp", 3},
    {oSETYCOMP, "setycomp", 3},
    {oSETZCOMP, "setzcomp", 3},    
    {oDU, "du", 1},
    {oDV, "dv", 1},
    {oDU3, "du3", 1},
    {oDV3, "dv3", 1},
    {oDERIV, "deriv", 1},
    {oDERIV3, "deriv3", 1},
    {oCELLNOISE, "cellnoise", 3},
    {oCELLNOISE3, "cellnoise3", 3},
    {oRANDOM, "random", 1},
    {oRANDOM3, "random3", 1},
    {oDEPTH, "depth", 1},
    {oPTLINED, "ptlined", 1},
    {oROTATE, "rotate", 1},
    {oFRESNEL, "fresnel", 1},
    {oFRESNEL2, "fresnel2", 1},
    {oFSTART, "fstart", 1},
    {oCALCULATENORMAL, "calculatenormal", 1},
    {oILLUM1, "illum1", 3},
    {oILLUM3, "illum3", 3},
    {oREFLECT, "reflect", 1},
    {oREFRACT, "refract", 1},
    {oARR, "arr", 1},
    {oARR3, "arr3", 1},
    {oCF2T, "cf2t", 1},
    {oPI, "pi", 1},
    {oPNOISE, "pnoise", 3},
    {oPNOISE3, "pnoise3", 3},
    {oCHK, "chk", 1},
    
};
#define NOPS    (sizeof(optab)/(sizeof(optab[0])))

struct ops *get_op(int op){
    int n;
    for (n=0; n<NOPS; n++){
        if (op == optab[n].op)
            return &optab[n];
    }
    /* invalid op. specified */
    yyerror("Code generation error: invalid op.\n");
    
    /*
        yyerror will invoke exit, so return is not needed, but to make compiler
        happy...
    */
    return NULL;
}

int is_branch_op(int op)
{

    switch (op){
        case oEQ3:
        case oEQ:
        case oEQS:
        case oNE3:
        case oNE:
        case oNES:
        case oLT:
        case oGT:
        case oLE:
        case oGE:
        case oJMP:
        case oILLUM1:
        case oILLUM3:
            return 1;
        default:
            return 0;
    }
}

int is_offset_op(int op)
{

    switch(op){
        case oASSIGN:
        case oASSIGN3:
        case oASSIGNS:
        case oADDEQ:
        case oADDEQ3:
        case oSUBEQ:
        case oSUBEQ3:
        case oMULEQ:
        case oMULEQ3:
        case oDIVEQ:
        case oDIVEQ3:
        case oPUSH:
        case oPUSH3:
        case oPUSHS:
        case oPUSHARR:
        case oPUSHARR3:
        case oSETCOMP:
        case oSETXCOMP:
        case oSETYCOMP:
        case oSETZCOMP:
        case oPUSHALEN:
        case oPRINTF:
            return 1;
            break;
        default:
            return 0;
    }
        
}


int is_string_op(int op)
{
    switch(op){
        case oASSIGNS:
        case oPUSHS:
        case oPRINTF:
            return 1;
            break;
        default:
            return 0;
    }
}

/*
    Remove function bodies from byte code and fix branch op. offsets
*/
void s_fixcode(unsigned char * code, int size)
{

    int offset=0;
    int counter=0;
    int current_op = code[counter];
    
    if (!output_functions){
        /* skip all previously declared functions by starting from shader */
        offset = counter = shaderstart;
        current_op = code[counter];
    }
        
    while(counter<size){
        if (is_branch_op(current_op)){
            short cur_offset = code[counter+2] | (code[counter+1]<<8);
            cur_offset -= offset;
            code[counter+2] = cur_offset&0xff;
            code[counter+1] = cur_offset>>8;
            counter += 3;
        }
        
        else if(current_op == oFSTART && !output_functions ){
            int func_add_offs = counter;
            int recursive_counter = 1;
            while (recursive_counter > 0){
                int copy_count = get_op(current_op)->size;
                counter += copy_count;
                current_op = code[counter];
                if ( current_op == oFSTART)
                    recursive_counter += 1;
                else if (current_op == oRET)
                    recursive_counter -= 1;
            }
            if (counter<size){
                counter+=get_op(current_op)->size;            
                offset += counter-func_add_offs;
            }
        }
        else{
            int copy_count = get_op(current_op)->size;
            counter += copy_count;
        }
        current_op = code[counter];
    }
}

/*
    Fix variable offsets (unused variables are removed from 
    variable stack)
*/
void s_fixvars(unsigned char * code, int size)
{

    int offset=0;
    int counter=0;
    int current_op;
    
    if (output_functions){
        return;
    }

    /* skip all previously declared functions by starting from shader */
    offset = counter = shaderstart;
    
    while(counter<size){
        int offset_val;
        current_op = code[counter];
        offset_val =is_offset_op(current_op);
        if (offset_val){
            short cur_offset = code[counter+offset_val+1] | 
                                  (code[counter+offset_val]<<8);
            short new_offset = relocate_sym(cur_offset);
            code[counter+offset_val+1] = new_offset&0xff;
            code[counter+offset_val] = new_offset>>8;
        }
        counter+=get_op(current_op)->size;            
    }
    end_sym_relocation();
}

/*
    dumps out opcodes as "assembler" code.
*/
void s_dumpobj(unsigned char *code, int size)
{
    int n = 0;
    int code_count = 0;
        
    if (!output_functions){
    /* skip all previously declared functions by starting from shader */
        n = shaderstart;
    }
    for(; n < size; n++) {
        int current_op = code[n];
        struct ops * current_opstr = get_op(current_op);
        if (current_opstr == NULL) {
            fprintf(stderr, "Unknown op@%d: %d\n", n, current_op);
        } 
        else if (!output_functions && current_op == oFSTART ){
            /* skip inline funcs */
            int recursive_counter = 1;
            while (recursive_counter > 0){
                int copy_count = get_op(current_op)->size;
                n += copy_count;
                current_op = code[n];
                if ( current_op == oFSTART)
                    recursive_counter += 1;
                else if (current_op == oRET)
                    recursive_counter -= 1;
            }
            /* oFRET will be skipped by for statement! */
        }                
        else {
            int k, l;
            int offset_val;
            fprintf(output_file,"# %04d:\t%-20s", code_count, current_opstr->name);
            if (current_opstr->size > 1)
                fputc('\t', output_file);
            for(k = 1; k < current_opstr->size; k += 2) {
                fprintf(output_file," %d", (code[n+k]<<8)|code[n+k+1]);
            }
            code_count += current_opstr->size;
            /* print additional info for vars and consts */
            offset_val = is_offset_op(current_op);
            if (offset_val){
                int offset = (code[n+offset_val]<<8)|code[n+offset_val+1];
                struct _sym *sym = getsym(offset);
                if (is_string_op(current_op)){
                    sym = get_string_sym(offset);
                }
                if( sym !=NULL){
                    if (sym->dim != 1){
                        /* print array element */
                        int index = offset-sym->offset;
                        if (istriple(sym->type)){
                            index /= 3;
                        }
                        fprintf(output_file,"\t%s[%1d]", sym->name, index);
                    }
                    else
                        fprintf(output_file,"\t%s", sym->name);
                }
                else if (is_constant(offset)){
                    if (is_string_op(current_op))
                        fprintf(output_file, "\t%s", getstring(offset));
                    else
                        fprintf(output_file,"\t<%f>", getfloat(offset));
                }
                else{
                    fprintf(stderr,"invalid offset @%d\n", n);
                }
            }
            else if (current_op == oFSTART){
                struct func_tree * f = get_function_by_offset(n);
                if (f){
                    fprintf(output_file,"\t\t%s", f->name);
                }
            }
            n += current_opstr->size - 1;
            fputc('\n', output_file);
        }
    }
}


int calc_codesize(unsigned char *code, int size)
{
    int n = 0;
    int code_count = 0;
        
    n = shaderstart;
    for(; n < size; n++) {
        int current_op = code[n];
        struct ops * current_opstr = get_op(current_op);
        if ( current_op == oFSTART ){
            /* skip inline funcs */
            int recursive_counter = 1;
            while (recursive_counter > 0){
                int copy_count = get_op(current_op)->size;
                n += copy_count;
                current_op = code[n];
                if ( current_op == oFSTART)
                    recursive_counter += 1;
                else if (current_op == oRET)
                    recursive_counter -= 1;
            }
            /* oFRET will be skipped by for statement! */
        }                
        else {
            code_count += current_opstr->size;
            n += current_opstr->size - 1;
        }
    }
    return code_count;
}



/* opcodes per line */
#define NROW    16
void s_dumpop(int op)
{
    static int counter = 1;
    fprintf(output_file," %3d", op);
        
    if (counter == NROW){
        fputc('\n', output_file);
        counter = 1;
    }
    else {
        counter++;
    }
}

/*
    dumps out opcodes as byte stream
*/
void s_dumpcode(unsigned char *code, int size)
{

    int n = shaderstart;
    
    fprintf(output_file,"-\n");
    if (output_functions){
        /* don't dump code in case of function output, as it will be invalid */
        return;
    }
    while ( n < size) {
        int cur_opcode = code[n];
        struct ops *cur_op = get_op( cur_opcode);
        if (cur_op == NULL){
            fprintf(stderr, "Unknown op @%d: %d\n", n, cur_opcode);
            n ++;
        }
        else if (cur_opcode == oFSTART && !output_functions ){
            /* skip inline funcs */
            int recursive_counter = 1;
            while (recursive_counter > 0){
                int copy_count = get_op(cur_opcode)->size;
                n += copy_count;
                cur_opcode = code[n];
                if ( cur_opcode == oFSTART)
                    recursive_counter += 1;
                else if (cur_opcode == oRET)
                    recursive_counter -= 1;
            }
            /* skip oFRET, assume op size 1 */
            n ++; 
        }                
        else {
            int op_size = cur_op->size;
            while (op_size >0){
                s_dumpop(code[n++]);
                op_size--;
            }
        }
    }
}


